#ifndef oid_hdr
#define oid_hdr

class PageKey;
class ObjectKey;
class ObjectId;

#include "odb.h"
#include "psi_term.h"

class ObjFile;

struct OId {
  ObjFile* file_d;
  vol_num vnum_d;
  page_num pnum_d;
  slot_num snum_d;
          // 0 values for pnum_d, snum_d indicate NULL values
  OId():file_d(NULL), vnum_d(0), pnum_d(0), snum_d(0) {}
};

class FileKey {

protected:
  ObjFile* file_d;
  vol_num vnum_d;			

public:
  FileKey(ObjFile* f):file_d(f), vnum_d(0) {}

  int operator==(const FileKey& f_key)const {
    return ( file_d==f_key.file_d);
  }

  vol_num xkey() const { return vnum_d; }

  ObjFile* file() const { return file_d; }
  const vol_num& vnum() const { return vnum_d; }

  friend ostream& operator<<(ostream& os, const FileKey& f_key);

protected:
  virtual void output_data(ostream& os)const;

};

////////////////////
// CLASS PageKey //
//////////////////
class PageKey:public FileKey {

protected:
  page_num pnum_d;		

public:
  PageKey(ObjFile* f, page_num p):FileKey(f), pnum_d(p) {}
  PageKey(const PageKey& p_key):FileKey(p_key.file_d),
                              pnum_d(p_key.pnum_d) {}
  PageKey(const OId& o_id):FileKey(o_id.file_d), pnum_d(o_id.pnum_d) {}

  int operator==(const PageKey& p_key)const {
    return ( file_d==p_key.file_d && pnum_d==p_key.pnum_d );
  }

  page_num ykey() const { return pnum_d; }

  const page_num& pnum() const { return pnum_d; }

protected:
  void output_data(ostream& os)const;

};

//////////////////////
// CLASS ObjectKey //
////////////////////
class ObjectKey:public PageKey {

protected:
  slot_num snum_d;
	  // 0 values for pnum_d, snum_d indicate NULL values

public:
  ObjectKey(const ObjectKey& o_key):PageKey(o_key), snum_d(o_key.snum_d)  {}
  ObjectKey(ObjFile* f, page_num p, slot_num s):PageKey(f,p), snum_d(s) {}
  ObjectKey(const OId& o_id):PageKey(o_id), snum_d(o_id.snum_d)  {}

  slot_num zkey() const { return snum_d; }
  
  const slot_num& snum() const { return snum_d; }

  int operator==(const ObjectKey& o_key)const {
    return ( file_d==o_key.file_d && pnum_d==o_key.pnum_d 
	    && snum_d==o_key.snum_d );
  }

protected:
  void output_data(ostream& os)const;

};

class PageObjectsNode;

/////////////////////
// CLASS ObjectId //
///////////////////
class ObjectId: public ObjectKey {
  friend class PageObjectsNode;
private:
  GENERIC dbpointer_d;
   // dbpointer_d points to database object location of the term
   // if in the database the psi_term corefs to sequence of psi_terms
   // dbpointer_d points to the final psi_term database location.
   // if dbpointer_d != deref_ptr(term_d)->dbpointer then the term
   // has been modified.

  // dbpointer_d maybe garbage unless the object_id is hashed. 
  // do not use it if object_id is not hashed. The gc function updates 
  // dbpointer_d through "ObjectNode-check_data function" only. The gc
  // doesnt seem to work when dbpointer_d is updated through term->dbpointer.

public:
  ObjectId(const ObjectKey& o_key, GENERIC dbp):
                                   ObjectKey(o_key), 
                                   dbpointer_d(dbp),
                                   state_d(0) {}

  void dbcoreffed_psi_term(GENERIC dbpointer) {
    dbpointer_d = dbpointer;
  }
  GENERIC dbpointer() { return dbpointer_d; }

  //  void oid_temp() {}

  void feature_added2psi_term() {
    state_d |= FEATURE_MASK;
  }
  void incomplete_object() {
    state_d |= INCOMPLETE_OBJ_MASK;
  }

  void updated_psi_term() {
    state_d &= ~MODIFIED_MASK;
  }
  int is_modified() {
    return (state_d & MODIFIED_MASK);
  }
  void hashed_object(int i=1) {
    if (i) state_d |= HASHED_MASK;
    else  {
      state_d &= ~HASHED_MASK;
    }
  }
  int is_hashed() {
    return (state_d & HASHED_MASK);
  }
  
  void check_data();

private:
  enum {FEATURE_MASK=1, INCOMPLETE_OBJ_MASK=2, HASHED_MASK=4};
  enum{MODIFIED_MASK=3};
  int state_d;
};


#endif
