#ifndef p_buffer_hdr
#define p_buffer_hdr

class Cache;
class PageNode;
class PageBuffer;

#include "odb.h"
#include "flist.h"
#include "oid.h"
#include "exc.h"

class ObjStream;

//////////////////
// CLASS Cache //
////////////////
class Cache  {
  
private:
  char loc_d[PAGE_SIZE];
  u_1byte state_d;
  int lock_d;
  int ptrs_lock_d;

  enum {DIRTYMASK=4};

public:
  Cache():state_d(0), lock_d(0), ptrs_lock_d(0) {}
  ~Cache() { }

  s_2byte size() const { return PAGE_SIZE; }

  operator char*() const { return (char*)loc_d; }
  char* operator[](int v) const {
    UNEXPECTED( (v >= PAGE_SIZE),
	       "in Cache:: operator[](int v), index out of range" );    
    return (char*)&(loc_d[v]); 
  }
  void operator()(s_2byte index, const char* d, s_2byte size);
  void operator()(s_2byte to, s_2byte from, s_2byte size);

  void lock();
  void un_lock();
  int is_locked() { 
    return (lock_d > 0) || (ptrs_lock_d > 0); 
  }
  int num_of_locks();

  void ptrs_lock(int y);
  int is_ptrs_locked() { 
    return ptrs_lock_d > 0; 
  }
  int num_of_ptr_locks();

  void dirty(int i) { 
    if (i) state_d |= DIRTYMASK;
    else state_d &= ~DIRTYMASK; 
  }
  int is_dirty() { return state_d & DIRTYMASK; }
  
private:
  Cache(const Cache& c);
  void operator=(const Cache& c);

};

/////////////////////
// CLASS PageNode //
///////////////////
class PageNode:public B_FL_Node {

public:
  static int max_num;

private:
  PageKey key_d;
  Cache cache_d;
  ObjPageInfo info_d;
  static GENERIC alloc_ptr;
  static int n_of_nodes_d;			// current num of PageNodes

public:
  PageNode(const PageKey& p_key):key_d(p_key)
  {
    UNEXPECTED( (this != (void*)BYTEDATA_DATA(alloc_ptr)),
	       "unexpected error in PageNode class" )
  }
  ~PageNode();
  void operator()(const PageKey& p_key) {
    key_d=p_key;
  }
  void flush_page(); // page is flushed if dirty. 
  const PageKey& key() const { return key_d; }
  Cache& cache() { return cache_d; }
  ObjPageInfo& info() { return info_d; }

  void* operator new(size_t s) {
    alloc_ptr = PsiTerm::byte_data_alloc(s);
    n_of_nodes_d ++;
    return (void*)BYTEDATA_DATA(alloc_ptr);
  }
  void operator delete(void *) { n_of_nodes_d--; }
  static int num_of_nodes() { return n_of_nodes_d; }

protected:

  void output_data(ostream&)const;
  
private:
  PageNode(const PageNode& n);  
  void operator=(const PageNode& n);

};

#include "hash_mat2.h"
typedef HashMatrix2<PageKey, PageNode> PageHashTable;

///////////////////////
// CLASS PageBuffer //
/////////////////////
class PageBuffer:public FList {
private:
  PageHashTable* table;

public:
  PageBuffer(int max_size,
	     int page_hash_size=0,  int file_hash_size=5);
  ~PageBuffer();

  PageNode* get_cache(const PageKey& p_key, int& state);
  // get_cache on a particular PageKey p_key, if the cache is not hashed
  // it calls new_cache
  PageNode* new_cache(const PageKey& p_key);
  // new_cache obtains unused cache, if not it flushes a used cache 
  // and returns it. The cache returned is hashed on p_key.
  void flush_pages();
  int evict_pages(int min_num=0);
  void evict_pages(ObjFile* file);

  REAL size() { return (PageNode::num_of_nodes())*sizeof(PageNode); }
  REAL max_size() { return (PageNode::max_num)*sizeof(PageNode); }
  int length() { 
    return PageNode::num_of_nodes(); 
  }
  int max_length() { return PageNode::max_num; }
  void check_data();
  
  void usage(ostream& os);
protected:
  PageNode* get_first_unlocked_node();
  PageNode* get_last_unlocked_node();
  PageNode* free_a();		
  // delete first unlocked node from buffer group
  PageNode* free_e();		
  // delete last unlocked node from buffer group
  void mv2start(PageNode* page_node) {
    page_node->remove2();
    prepend(page_node);    
  }
  // page node shifted to start of buffer group.
  void mv2end(PageNode* page_node) {
    page_node->remove2();
    append(page_node);
  }
  // page node shifted to end of buffer group.
  PageNode* first_page() {
    return (PageNode*) first();
  }
  PageNode* next_page(PageNode* page_node) {
    return (PageNode*)page_node->next2();
  }
  PageNode* prev_page(PageNode* page_node) {
    return (PageNode*)page_node->prev2();
  }
  //  int buf_length();

private:
  PageBuffer(const PageBuffer& n);
  void operator=(const PageBuffer& n);
};

#endif
